/*
  Projekt: Network Driver for NIC 3c509
           TUOX Projekt
  Autori:  Miroslav Kubiczek
           Martin Fanta
*/
#include <time.h>
#include <stdlib.h>
#include <mem.h>
#include <stdio.h>
#include <conio.h>
#include <dos.h>
#include <alloc.h>

#include "stuff.h"
#include "driver.h"

static char *version = "3c509 driver for TUOX\n";

#define ENODEV 1

#define IFF_ALLMULTI 1
#define IFF_PROMISC  2

#ifdef EL3_DEBUG
int el3_debug = EL3_DEBUG;
#else
int el3_debug = 6;
#endif

static int id_port = 0x100;
static ushort id_read_eeprom(int index);
static int el3_open(struct device *dev);
int el3_probe(struct device *dev);
static int el3_start_xmit(struct device *dev, uchar *data, uint len);
static void el3_interrupt(struct device *dev);
static int el3_rx(struct device *dev);
static int el3_close(struct device *dev);

/*
  Inicializuj adapter (pouze pro ISA sbernici).
  Poznamka: Tato funkce musi probihat v cyklu, dokud nase sitova karta
            neni nalezena.
            Podle teorie chaosu - jednou se prece musime trefit.
*/
int el3_probe(struct device *dev) {
  short lrs_state, i;
  ushort ioaddr, irq, if_port;
  short *phys_addr = (short *)dev->dev_addr;
  unsigned short iobase = id_read_eeprom(8);
  const char *if_names[] = {"10baseT", "AUI", "undefined", "BNC"};

  static int current_tag = 0;

  /* Reset the ISA PnP activation mechanism on 3c509b. */
  outb(0x02, 0x279);           /* Select PnP config control register. */
  outb(0x02, 0xA79);           /* Return to WaitForKey state. */


  /* Standard ISA s activatio mechanism on 3c509 */
	/* Select an open I/O port location at 0x1*0 to do contention select. */
  /* (Simply select I/O port to communicate
      with devices connected to ISA bus) */

  for (id_port = 0x100; id_port < 0x200; id_port += 0x10) {
    outb(0x00, id_port);
    outb(0xff, id_port);
    if (inb(id_port) & 0x01) break;
  }

  if (id_port >= 0x200) {             /* GCC optimizes this test out. */
    /* Rare -- do we really need a warning? */
    printf(" WARNING: No I/O port available for 3c509 activation.\n");
    return -ENODEV;
  }

  /* Next check for all ISA bus boards by sending the ID sequence to the
     ID_PORT.  We find cards past the first by setting the 'current_tag'
	   on cards as they are found.  Cards with their tag set will not
	   respond to subsequent ID sequences. */
  /* Choose the port */
  outb(0x00, id_port);
  outb(0x00, id_port);
  /* Write special ISA IDS-sequence so that state changed
     from ID_WAIT to ID_CMD (command) */
  lrs_state = 0xFF;
  for(i = 0; i < 255; i++) {
    outb(lrs_state, id_port);
    lrs_state <<= 1;
    lrs_state = lrs_state & 0x100 ? lrs_state ^ 0xcf : lrs_state;
  }
  /* current_tag == 0 in first use this rutine
     first cycle of "probe" all NIC 3c509 on ISA */
  if (current_tag == 0)
    /* For the first probe, clear all board's tag registers.
       Marked adapter's tag register means that we have found it but
       we do not need this one so that we will kill it. */
    outb(0xd0, id_port);
  else
    /* Otherwise kill off already-found boards.
       (They will not be responding) */
    outb(0xd8, id_port);

	/* Select the one adapter but only on the ISA bus
     ==> Read in EEPROM data, which does contention-select.
	       Only the lowest address board will stay "on-line".
	       3Com got the byte order backwards. */
  /* Read Manufacture ID */
  if (id_read_eeprom(7) != 0x6d50) {
    /* We have not found any 3com5x9 card
       but another adapter on ISA */
    return -ENODEV;
  }
  /* Read physical address of our NIC (Network Interface Card) */
  for (i = 0; i < 3; i++) phys_addr[i] = htons(id_read_eeprom(i));

  /* The Next can be read directly from 3com5x9' WINDOW 0 registers
     after the adapter is activated, it is easier but not so funny.
     The Address Configuration, Resource Configuration and Product ID are
     copied after reset. */
  /* Read Address Configuration EEPROM Data */
  iobase = id_read_eeprom(8);
    /* Transceiver Type Select in bits 15,14
       00 TP (10BASE-T)
       01 AUI
       11 BNC */
  if_port = iobase >> 14;
    /* I/O Ports of NIC
       ioaddr = (Bits 5-0) * 16d + 200h */
  ioaddr = 0x200 + ((iobase & 0x1f) << 4);
  /* Set interrupt */
  if (dev->irq > 1  &&  dev->irq < 16) irq = dev->irq;
  else
    /* Read Resource Configuration EEPROM Data to find IRQ number */
    irq = id_read_eeprom(9) >> 12;

    /* Set the adaptor tag so that the next card can be found. */
  outb(0xd0 + ++current_tag, id_port);

  /* Activate the adaptor at the EEPROM location. */
  /* Write 0xFF activates the adapter
     at the preconfigured I/O base address */
  outb((ioaddr >> 4) | 0xe0, id_port);

  /* Test adapter I/O ports location and conflicts */
  EL3WINDOW(0);
  if (inw(ioaddr) != 0x6d50) return -ENODEV;

  /* Free the interrupt so that some other card can use it.
     Disable all IRQ line drivers
  */
  outw(0x0f00, ioaddr + WN0_IRQ);

  /* Set system's structures */
  dev->base_addr = ioaddr;
  dev->irq = irq;
  dev->if_port = if_port;

  /* Show some very useful information */
  printf( "%s: 3c509 at %#3.3x tag %d, %s port, address ",
    dev->name, dev->base_addr, current_tag, if_names[dev->if_port]);
	/* Read in the station address. */
  for (i = 0; i < 6; i++) printf("%2.2x", dev->dev_addr[i]);
  printf(", IRQ %d.\n", dev->irq);

  if (el3_debug > 0) printf(version);

  /* Fill the EL3-specific entries in the device structure. */
  dev->open = &el3_open;
  dev->hard_start_xmit = &el3_start_xmit;
  dev->close = &el3_close;
  dev->t_interrupt = &el3_interrupt;

  return 0;
}

/*
  Read a word from the EEPROM when in the ISA ID probe state.
*/
static ushort id_read_eeprom(int index) {
  int bit, word = 0;

  /* Issue read command, and pause for at least 162 us. for it to complete.
      Assume extra-fast 16Mhz bus. */
  outb(EEPROM_READ + index, id_port);

  /* Pause for at least 162 us. for the read to take place. */
  udelay (20);
  for (bit = 15; bit >= 0; bit--) word = (word << 1) + (inb(id_port) & 0x01);
  if (el3_debug > 3) printf("  3c509 EEPROM word %d %#4.4x.\n", index, word);

  return word;
}

/*
  Pocatecni inicializace NIC. Je zavisla na systemu
  a na jeho sitovych ovladacich.
*/
static int el3_open(struct device *dev) {
  int ioaddr = dev->base_addr;
  int i;

  /* Reset tranciever and reciever */
  outw(TxReset, ioaddr + EL3_CMD);
  outw(RxReset, ioaddr + EL3_CMD);
  outw(SetStatusEnb | 0x00, ioaddr + EL3_CMD);

  /* Set irq for the card, system routine */
  enable_irq(dev->irq);

  EL3WINDOW(0);
  if (el3_debug > 3)
    printf("%s: Opening, IRQ %d	 status@%x %4.4x.\n", dev->name,
      dev->irq, ioaddr + EL3_STATUS, inw(ioaddr + EL3_STATUS));

  /* Activate board: this is probably unnecessary. */
  outw(0x0001, ioaddr + 4);

  /* Set the IRQ line to number stored in dev */
  outw((dev->irq << 12) | 0x0f00, ioaddr + WN0_IRQ);

  /* Set the station address in window 2 each time opened. */
  EL3WINDOW(2);
  for (i = 0; i < 6; i++) outb(((uchar *)dev->dev_addr)[i], ioaddr + i);

  /* Action specifics for transciever media type
       0 TP (10BASE-T)
       1 AUI
       3 BNC  */
  if (dev->if_port == 3)
    /* Start the thinnet transceiver. We should really wait 50ms...*/
    outw(StartCoax, ioaddr + EL3_CMD);
  else if (dev->if_port == 0) {
    /* 10baseT interface, enabled link beat and jabber check. */
    EL3WINDOW(4);
    outw(inw(ioaddr + WN4_MEDIA) | MEDIA_TP, ioaddr + WN4_MEDIA);
  }

  /* Switch to the statistic window, and clear all stats by reading. */
  outw(StatsDisable, ioaddr + EL3_CMD);
  EL3WINDOW(6);
  for (i = 0; i < 9; i++)	inb(ioaddr + i);
  inw(ioaddr + 10);
  inw(ioaddr + 12);

  /* Switch to window 1 for normal use. */
  EL3WINDOW(1);

  /* Accept b-case and phys addr only. */
  outw(SetRxFilter | RxStation | RxBroadcast, ioaddr + EL3_CMD);
  /* Turn on statistics. */
  outw(StatsEnable, ioaddr + EL3_CMD);

  /* for System */
  dev->inter = 0;
  dev->tbusy = 0;
  dev->start = 1;

  outw(RxEnable, ioaddr + EL3_CMD); /* Enable the receiver. */
  outw(TxEnable, ioaddr + EL3_CMD); /* Enable transmitter. */
  /* Allow status bits to be seen. */
  outw(SetStatusEnb | 0xff, ioaddr + EL3_CMD);
  /* Acknowledge all pending events */
  outw(AckIntr | IntLatch | TxAvailable | RxEarly | IntReq,
    ioaddr + EL3_CMD);
  /* and set active indicator's and interrupt's mask. */
  outw(SetIntrEnb | IntLatch | TxAvailable | RxComplete | StatsFull,
    ioaddr + EL3_CMD);

  if (el3_debug > 3) printf("%s: Opened 3c509  IRQ %d  status %4.4x.\n",
		       dev->name, dev->irq, inw(ioaddr + EL3_STATUS));

  return 0;					/* Always succeed */
}

/*
  Zacni vysilat dany paket.

  Vysilani probiha nasledujicim zpusobem:
    Nejprve testuj dev->tbusy.
      Je nastavena  - prechozi paket jeste nebyl odeslan.
	Pokud je to v norme (casove) vrat se, jinak
	vrat chybu.
      Neni nastavena -
	Pridavej data paketu do Tx FIFO zapisovanim na porty.
	Vysilani zacne, az je zapsan cely paket (mame to nastavene takto).
	Jeste testuj zbyle misto v FIFO, pokud ho neni dostatek,
	nastav dev->tbusy a take umozni preruseni, ktere bude
	indikovat, volne misto v FIFO.
	Nakonec vycisti zasobnik stavu predchozich pakety

  Pozn: Delka len je v bytech.
*/
static int el3_start_xmit(struct device *dev, uchar *data, uint len) {
  uint tx_free;
  int ioaddr = dev->base_addr;

  /* Transmitter timeout, serious problems. */
  if (dev->tbusy) {
    int ticksofar = jiffies - dev->trans_start;
    if (ticksofar < 40)
    return 1;
				   /* presahla 40 s */
    printf("%s: transmit timed out, Tx_status %2.2x status %4.4x "
      "Tx FIFO room %d.\n",
      dev->name, inb(ioaddr + TX_STATUS), inw(ioaddr + EL3_STATUS),
      inw(ioaddr + TX_FREE));
    dev->trans_start = jiffies;
    /* External Error - Reset for media type TP */
    /* Issue TX_RESET and TX_START commands. */
    outw(TxReset, ioaddr + EL3_CMD);
    outw(TxEnable, ioaddr + EL3_CMD);
    dev->tbusy = 0;
  }

  /* No input */
  if (data == NULL) return 0;
  if (len <= 0) return 0;

  if (el3_debug > 3)
    printf("%s: el3_start_xmit(length = %u) called, status %4.4x.\n",
      dev->name, len, inw(ioaddr + EL3_STATUS));


#ifdef TEST_ERROR_TX
  {	/* Error-checking code, delete someday. */
    ushort status = inw(ioaddr + EL3_STATUS);
    if (status & 0x0001 		/* IRQ line active, missed one. */
      && inw(ioaddr + EL3_STATUS) & 1) { 			/* Make sure. */
      printf("%s: Missed interrupt, status then %04x now %04x"
	"  Tx %2.2x Rx %4.4x.\n", dev->name, status,
	inw(ioaddr + EL3_STATUS), inb(ioaddr + TX_STATUS),
	inw(ioaddr + RX_STATUS));
      /* Fake interrupt trigger by masking, acknowledge interrupts. */
      outw(SetStatusEnb | 0x00, ioaddr + EL3_CMD);
      outw(AckIntr | IntLatch | TxAvailable | RxEarly | IntReq,
	ioaddr + EL3_CMD);
      outw(SetStatusEnb | 0xff, ioaddr + EL3_CMD);
    }
  }
#endif

  /* Start with writing packet to FIFO */
  /* Put out the doubleword header... */
  outw(len, ioaddr + TX_FIFO);
  outw(0x00, ioaddr + TX_FIFO);
  /* ... and the packet rounded to a doubleword. */
  outsl(ioaddr + TX_FIFO, data, (len + 3) >> 2);

  dev->trans_start = jiffies;

  /* Number of free bytes in FIFO */
  tx_free = inw(ioaddr + TX_FREE);

  if (el3_debug > 3) printf("TX_FREE %u Muj thrashold ... \n", tx_free);
  if (tx_free > 1536) {
    dev->tbusy = 0;
  }
  else {
    // Interrupt us when the FIFO has room for max-sized packet.
    outw(SetTxThreshold + 1536, ioaddr + EL3_CMD);
    dev->tbusy = 1;
  }

  /* Clear the Tx status stack. */
  {
    short tx_status;
    int i = 4;

    while (--i>0 && (tx_status = inb(ioaddr + TX_STATUS)) > 0) {
      printf("CLEAR TX STATUS\n");
      if (tx_status & 0x30) outw(TxReset, ioaddr + EL3_CMD);
      if (tx_status & 0x3C) outw(TxEnable, ioaddr + EL3_CMD);
      outb(0x00, ioaddr + TX_STATUS); /* Pop the status stack. */
    }
  }
  return 0;
}

/*
  Obsluz preruseni NIC.
  Pouze:
       TxAvailable - dany pocet bytu je uz minimalne volnych v trc. FIFO
                     (znovu ho povol).
       RxComplete  - prijmy paket
       StatsFull   - preplne statisticke registry

  Dalsi preruseni nejsou normalne povolene:
       AdapterFailure - chyba karty (pri prijmu ci neco jineho)
       RxEarly     - prisel dany pocet bytu

  Na konci preruseni potvrd obsluhu daneho preruseni (Acknowledge)
*/
static void el3_interrupt(struct device *dev) {
  int ioaddr, status;
  int i = 0;

  if (dev->inter)
    printf("%s: Re-entering the interrupt handler.\n", dev->name);

  dev->inter = 1;

  ioaddr = dev->base_addr;
  status = inw(ioaddr + EL3_STATUS);

  if (el3_debug > 2)
    printf("%s: interrupt, status %4.4x.\n", dev->name, status);

  while ((status = inw(ioaddr + EL3_STATUS)) &
	(IntLatch | RxComplete | StatsFull)) {

    /* Packet received */
    if (status & RxComplete) el3_rx(dev);

    /* Free Space in trc. FIFO */
    if (status & TxAvailable) {
      if (el3_debug > 5) printf("interrupt: TX room bit was handled.\n");
	/* There's room in the FIFO for a full-sized packet. */
	outw(AckIntr | TxAvailable, ioaddr + EL3_CMD);
	dev->tbusy = 0;
      }

      /* Handle all uncommon interrupts. */
      if (status & (AdapterFailure | RxEarly | StatsFull)) {
	if (status & RxEarly) {		    /* Rx early is unused. */
	  el3_rx(dev);
	  outw(AckIntr | RxEarly, ioaddr + EL3_CMD);
	}
      if (status & AdapterFailure) {
	/* Adapter failure requires Rx reset and reinit. */
	outw(RxReset, ioaddr + EL3_CMD);
	/* Set the Rx filter to the current state. */
	outw(SetRxFilter | RxStation | RxBroadcast
	  | (dev->flags & IFF_ALLMULTI ? RxMulticast : 0)
	  | (dev->flags & IFF_PROMISC ? RxProm : 0),
	  ioaddr + EL3_CMD);
	outw(RxEnable, ioaddr + EL3_CMD); /* Re-enable the receiver. */
	outw(AckIntr | AdapterFailure, ioaddr + EL3_CMD);
      }
    }

    if (++i > 10) {
      printf("%s: Infinite loop in interrupt, status %4.4x.\n",
      dev->name, status);
      /* Clear all interrupts. */
      outw(AckIntr | 0xFF, ioaddr + EL3_CMD);
      break;
    }
    /* Acknowledge the IRQ. */
    outw(AckIntr | IntReq | IntLatch, ioaddr + EL3_CMD); /* Ack IRQ */
  }

  if (el3_debug > 4) {
    printf("%s: exiting interrupt, status %4.4x.\n", dev->name,
      inw(ioaddr + EL3_STATUS));
  }

  dev->inter = 0;
  return;
}

/*
  Prisel paket,
  ted  nacti stav paketu a pokud neni chyba,
  nacti data z receiver FIFO.
  Nakonec vzdy potrvrd prijem paketu a odstran ho z FIFO
  pomoci RxDiscard. Tim ho zaroven zapoctes pro statistiku.
*/
static int el3_rx(struct device *dev) {
  int ioaddr = dev->base_addr;
  short rx_status;

  if (el3_debug > 5)
    printf("   In rx_packet(), status %4.4x, rx_status %4.4x.\n",
      inw(ioaddr+EL3_STATUS), inw(ioaddr+RX_STATUS));

  while ((rx_status = inw(ioaddr + RX_STATUS)) > 0) {
    /* Bit 14 means error in packet but not incomplete packet! */
    if (rx_status & 0x4000) {
      /* Error, only update statics registers */
      printf("el3_rx: error rx_status 0x%x !\n", rx_status & 0x3800);
    }
    else {
      /* Without error rx_status is packet length */
      short pkt_len = rx_status & 0x7ff;
      uchar *muj_paket;

      /* Pri prijmu vypisu dany paket */
      muj_paket = (uchar *)malloc(pkt_len+5);
      memset(muj_paket, '*',pkt_len+5);

      if (el3_debug > 4)
	printf( "rx: Receiving packet size %d status %4.4x.\n",
	  pkt_len, rx_status);

      if (muj_paket != NULL) {
	int i;

	/* Read packet data, pkt_len is in bytess,
	   hence convert it to dword */
	insl(ioaddr+RX_FIFO, muj_paket, (pkt_len + 3) >> 2);
	for (i=0; i<pkt_len; i++)
	  // if (muj_paket[i]>=32)
	  printf("%x#",muj_paket[i]);
	printf("\n");
	free(muj_paket);

	outw(RxDiscard, ioaddr + EL3_CMD); /* Pop top Rx packet. */
	continue;

      }
      else if (el3_debug)
	printf("%s: el3_rx :Couldn't allocate a buff of size %d.\n",
	  dev->name, pkt_len);
    }

    /* Pop top Rx packet and update statistics */
    outw(RxDiscard, ioaddr + EL3_CMD);
    while (inw(ioaddr + EL3_STATUS) & 0x1000)
      printf("	Waiting for 3c509 to discard packet, status %x.\n",
	inw(ioaddr + EL3_STATUS) );
  }

  return 0;
}

/*
  Ukonci praci s NIC:
  konec se statistikou, prijmem a vysilanim paketu,
  komunikaci na lince a hlavne s prerusenim.
*/
static int el3_close(struct device *dev) {
  int ioaddr = dev->base_addr;

  if (el3_debug > 2) printf("%s: Shutting down ethercard.\n", dev->name);

  dev->tbusy = 1;
  dev->start = 0;

  /* Turn off statistics ASAP.  We update lp->stats below. */
  outw(StatsDisable, ioaddr + EL3_CMD);

  /* Disable the receiver and transmitter. */
  outw(RxDisable, ioaddr + EL3_CMD);
  outw(TxDisable, ioaddr + EL3_CMD);

  /* Action specics for different media type */
  if (dev->if_port == 3)
    /* Turn off thinnet power.  Green! */
    outw(StopCoax, ioaddr + EL3_CMD);
  else if (dev->if_port == 0) {
    /* Disable link beat and jabber, if_port may change ere next open(). */
    EL3WINDOW(4);
    outw(inw(ioaddr + WN4_MEDIA) & ~MEDIA_TP, ioaddr + WN4_MEDIA);
  }

  /* Switching back to window 0 disables the IRQ. */
  EL3WINDOW(0);
  /* But we explicitly zero the IRQ line select anyway. */
  outw(0x0f00, ioaddr + WN0_IRQ);

  return 0;
}
